@@ -2,7 +2,7 @@
 #
 # test parsing of various resize arguments
 
-source "$TOP/tests/common"
+source "$TEST_TOP/common"
 
 check_prereq mkfs.btrfs
 check_prereq btrfs
@@ -10,7 +10,7 @@ check_prereq btrfs
 setup_root_helper
 prepare_test_dev 2g
 
-run_check "$TOP/mkfs.btrfs" -f "$IMAGE"
+run_check_mkfs_test_dev
 run_check_mount_test_dev
 
 # missing the one of the required arguments
